/*
  Stockfish, a UCI chess playing engine derived from Glaurung 2.1
  Copyright (C) 2004-2008 Tord Romstad (Glaurung author)
  Copyright (C) 2008-2015 Marco Costalba, Joona Kiiski, Tord Romstad
  Copyright (C) 2015-2019 Marco Costalba, Joona Kiiski, Gary Linscott, Tord Romstad

  Stockfish is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Stockfish is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <thread>

#include "bitboard.h"
#include "position.h"
#include "search.h"
#include "thread.h"
#include "tt.h"
#include "uci.h"
#include "syzygy/tbprobe.h"

#include <grpc/grpc.h>
#include <grpc++/server.h>
#include <grpc++/server_builder.h>
#include "hashprobe.grpc.pb.h"

using grpc::Server;
using grpc::ServerBuilder;
using grpc::ServerContext;
using grpc::Status;
using grpc::StatusCode;
using namespace hashprobe;

class HashProbeImpl final : public HashProbe::Service {
public:
	Status Probe(ServerContext* context,
	             const HashProbeRequest* request,
	             HashProbeResponse *response) {
		Position pos(request->fen(), /*isChess960=*/false, Threads.main());
		if (!pos.pos_is_ok()) {
			return Status(StatusCode::INVALID_ARGUMENT, "Invalid FEN");
		}

		bool invert = (pos.side_to_move() == BLACK);
		Search::StateStackPtr setup_states = Search::StateStackPtr(new std::stack<StateInfo>);

		ProbeMove(&pos, setup_states.get(), invert, response->mutable_root());

		MoveList<LEGAL> moves(pos);
		for (const ExtMove* em = moves.begin(); em != moves.end(); ++em) {
			HashProbeLine *line = response->add_line();
			FillMove(em->move, line->mutable_move());
			setup_states->push(StateInfo());
			pos.do_move(em->move, setup_states->top(), pos.gives_check(em->move, CheckInfo(pos)));
			ProbeMove(&pos, setup_states.get(), !invert, line);
			pos.undo_move(em->move);
		}

		return Status::OK;
	}

	void FillMove(Move move, HashProbeMove* decoded) {
		if (!is_ok(move)) return;

		Square from = from_sq(move);
		Square to = to_sq(move);

		if (type_of(move) == CASTLING) {
			to = make_square(to > from ? FILE_G : FILE_C, rank_of(from));
		}
			
		decoded->set_from_sq(UCI::square(from));
		decoded->set_to_sq(UCI::square(to));

		if (type_of(move) == PROMOTION) {
			decoded->set_promotion(std::string() + " PNBRQK"[promotion_type(move)]);
		}
	}

	void ProbeMove(Position* pos, std::stack<StateInfo>* setup_states, bool invert, HashProbeLine* response) {
		bool found;
		TTEntry *entry = TT.probe(pos->key(), found);
		response->set_found(found);
		if (found) {
			Value value = entry->value();
			Value eval = entry->eval();
			Bound bound = entry->bound();

			if (invert) {
				value = -value;
				eval = -eval;
				if (bound == BOUND_UPPER) {
					bound = BOUND_LOWER;
				} else if (bound == BOUND_LOWER) {
					bound = BOUND_UPPER;
				}
			}

			response->set_depth(entry->depth());
			FillValue(eval, response->mutable_eval());
			if (entry->depth() > DEPTH_NONE) {
				FillValue(value, response->mutable_value());
			}
			response->set_bound(HashProbeLine::ValueBound(bound));

			// Follow the PV until we hit an illegal move.
			std::stack<Move> pv;
			std::set<Key> seen;
			while (found && is_ok(entry->move())) {
				FillMove(entry->move(), response->add_pv());
				if (seen.count(pos->key())) break;
				pv.push(entry->move());
				seen.insert(pos->key());
				setup_states->push(StateInfo());
				pos->do_move(entry->move(), setup_states->top(), pos->gives_check(entry->move(), CheckInfo(*pos)));
				entry = TT.probe(pos->key(), found);
			}

			// Unroll the PV back again, so the Position object remains unchanged.
			while (!pv.empty()) {
				pos->undo_move(pv.top());
				pv.pop();
			}
		}
	}

	void FillValue(Value value, HashProbeScore* score) {
		if (abs(value) < VALUE_MATE - MAX_PLY) {
			score->set_score_type(HashProbeScore::SCORE_CP);
			score->set_score_cp(value * 100 / PawnValueEg);
		} else {
			score->set_score_type(HashProbeScore::SCORE_MATE);
			score->set_score_mate((value > 0 ? VALUE_MATE - value + 1 : -VALUE_MATE - value) / 2);
		}
	}
};

class HashProbeThread {
public:
	HashProbeThread(const std::string &server_address) {
		builder.AddListeningPort(server_address, grpc::InsecureServerCredentials());
		builder.RegisterService(&service);
		server = std::move(builder.BuildAndStart());
		std::cout << "Server listening on " << server_address << std::endl;
		std::thread([this]{ server->Wait(); }).detach();
	}

	void Shutdown() {
		server->Shutdown();
	}

private:
	HashProbeImpl service;
	ServerBuilder builder;
	std::unique_ptr<Server> server;
};

namespace PSQT {
  void init();
}

int main(int argc, char* argv[]) {

  std::cout << engine_info() << std::endl;

  UCI::init(Options);
  PSQT::init();
  Bitboards::init();
  Position::init();
  Bitbases::init();
  Search::init();
  Pawns::init();
  Threads.set(Options["Threads"]);
  Search::clear(); // After threads are up

  HashProbeThread thr("0.0.0.0:50051");

  UCI::loop(argc, argv);

  Threads.set(0);
  return 0;
}
